@@ -22,11 +22,13 @@ impl InMemoryDB {
 /// Memory backend, storing all state values in a `Map` in memory.
 #[derive(Debug, Clone)]
 pub struct CacheDB<ExtDB: DatabaseRef> {
-    /// dummy account info where code is allways None. Code bytes can be found in `contracts`
+    /// Dummy account info where `code` is always `None`.
+    /// Code bytes can be found in `contracts`.
     cache: Map<H160, AccountInfo>,
     storage: Map<H160, Map<U256, U256>>,
     contracts: Map<H256, Bytes>,
     logs: Vec<Log>,
+    block_hashes: Map<U256, H256>,
     db: ExtDB,
 }
 
@@ -40,6 +42,7 @@ impl<ExtDB: DatabaseRef> CacheDB<ExtDB> {
             storage: Map::new(),
             contracts,
             logs: Vec::default(),
+            block_hashes: Map::new(),
             db,
         }
     }
@@ -47,6 +50,7 @@ impl<ExtDB: DatabaseRef> CacheDB<ExtDB> {
     pub fn cache(&self) -> &Map<H160, AccountInfo> {
         &self.cache
     }
+
     pub fn storage(&self) -> &Map<H160, Map<U256, U256>> {
         &self.storage
     }
@@ -71,7 +75,7 @@ impl<ExtDB: DatabaseRef> CacheDB<ExtDB> {
     }
 }
 
-/// TODO it is currently only commiting to cached in memory db
+// TODO It is currently only commiting to cached in-memory DB
 impl<ExtDB: DatabaseRef> DatabaseCommit for CacheDB<ExtDB> {
     fn commit(&mut self, changes: Map<H160, Account>) {
         for (add, acc) in changes {
@@ -101,7 +105,14 @@ impl<ExtDB: DatabaseRef> DatabaseCommit for CacheDB<ExtDB> {
 
 impl<ExtDB: DatabaseRef> Database for CacheDB<ExtDB> {
     fn block_hash(&mut self, number: U256) -> H256 {
-        self.db.block_hash(number)
+        match self.block_hashes.entry(number) {
+            Entry::Occupied(entry) => *entry.get(),
+            Entry::Vacant(entry) => {
+                let hash = self.db.block_hash(number);
+                entry.insert(hash);
+                hash
+            }
+        }
     }
 
     fn basic(&mut self, address: H160) -> AccountInfo {
@@ -117,7 +128,9 @@ impl<ExtDB: DatabaseRef> Database for CacheDB<ExtDB> {
         }
     }
 
-    /// Account is assumed that is already loaded, just get storage slot from db.
+    /// Get the value in an account's storage slot.
+    ///
+    /// It is assumed that account is already loaded.
     fn storage(&mut self, address: H160, index: U256) -> U256 {
         match self.storage.entry(address) {
             Entry::Occupied(mut entry) => match entry.get_mut().entry(index) {
@@ -149,12 +162,44 @@ impl<ExtDB: DatabaseRef> Database for CacheDB<ExtDB> {
     }
 }
 
+impl<ExtDB: DatabaseRef> DatabaseRef for CacheDB<ExtDB> {
+    fn block_hash(&self, number: U256) -> H256 {
+        match self.block_hashes.get(&number) {
+            Some(entry) => *entry,
+            None => self.db.block_hash(number),
+        }
+    }
+
+    fn basic(&self, address: H160) -> AccountInfo {
+        match self.cache.get(&address) {
+            Some(info) => info.clone(),
+            None => self.db.basic(address),
+        }
+    }
+
+    fn storage(&self, address: H160, index: U256) -> U256 {
+        match self.storage.get(&address) {
+            Some(entry) => match entry.get(&index) {
+                Some(entry) => *entry,
+                None => self.db.storage(address, index),
+            },
+            None => self.db.storage(address, index),
+        }
+    }
+
+    fn code_by_hash(&self, code_hash: H256) -> Bytes {
+        match self.contracts.get(&code_hash) {
+            Some(entry) => entry.clone(),
+            None => self.db.code_by_hash(code_hash),
+        }
+    }
+}
+
+/// An empty database that always returns default values when queried.
 #[derive(Debug, Default, Clone)]
 pub struct EmptyDB();
 
 impl DatabaseRef for EmptyDB {
-    /// Whether account at address exists.
-    //fn exists(&self, address: H160) -> Option<AccountInfo>;
     /// Get basic account information.
     fn basic(&self, _address: H160) -> AccountInfo {
         AccountInfo::default()
@@ -169,18 +214,20 @@ impl DatabaseRef for EmptyDB {
     }
 
     // History related
-    fn block_hash(&self, _number: U256) -> H256 {
-        H256::default()
+    fn block_hash(&self, number: U256) -> H256 {
+        let mut buffer: [u8; 4 * 8] = [0; 4 * 8];
+        number.to_big_endian(&mut buffer);
+        H256::from_slice(&Keccak256::digest(&buffer))
     }
 }
 
-/// Very custom and dumm struct that will return accoun on address 0x0
+/// Custom benchmarking DB that only has account info for the zero address.
+///
+/// Any other address will return an empty account.
 #[derive(Debug, Default, Clone)]
 pub struct BenchmarkDB(pub Bytes);
 
 impl Database for BenchmarkDB {
-    /// Whether account at address exists.
-    //fn exists(&self, address: H160) -> Option<AccountInfo>;
     /// Get basic account information.
     fn basic(&mut self, address: H160) -> AccountInfo {
         if address == H160::zero() {
@@ -193,10 +240,12 @@ impl Database for BenchmarkDB {
         }
         AccountInfo::default()
     }
+
     /// Get account code by its hash
     fn code_by_hash(&mut self, _code_hash: H256) -> Bytes {
         Bytes::default()
     }
+
     /// Get storage value of address at index.
     fn storage(&mut self, _address: H160, _index: U256) -> U256 {
         U256::default()
