@@ -1,14 +1,14 @@
 use crate::{
     db::Database,
-    instructions::gas,
-    machine,
-    machine::{Contract, Gas, Machine},
+    gas, interpreter,
+    interpreter::{Contract, Interpreter},
     models::SelfDestructResult,
     return_ok,
-    spec::{Spec, SpecId::*},
     subroutine::{Account, State, SubRoutine},
-    util, CallContext, CreateScheme, Env, Inspector, Log, Return, TransactOut, TransactTo,
-    Transfer, KECCAK_EMPTY,
+    CallContext, CallInputs, CallScheme, CreateInputs, CreateScheme, Env, Gas, Inspector, Log,
+    Return, Spec,
+    SpecId::*,
+    TransactOut, TransactTo, Transfer, KECCAK_EMPTY,
 };
 use alloc::vec::Vec;
 use bytes::Bytes;
@@ -116,24 +116,33 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> Transact
                 let context = CallContext {
                     caller,
                     address,
+                    code_address: address,
                     apparent_value: value,
+                    scheme: CallScheme::Call,
                 };
-                let (exit, gas, bytes) = self.call_inner::<GSPEC>(
-                    address,
-                    Transfer {
+                let mut call_input = CallInputs {
+                    contract: address,
+                    transfer: Transfer {
                         source: caller,
                         target: address,
                         value,
                     },
-                    data,
+                    input: data,
                     gas_limit,
                     context,
-                );
+                };
+                let (exit, gas, bytes) = self.call_inner::<GSPEC>(&mut call_input);
                 (exit, gas, TransactOut::Call(bytes))
             }
             TransactTo::Create(scheme) => {
-                let (exit, address, ret_gas, bytes) =
-                    self.create_inner::<GSPEC>(caller, scheme, value, data, gas_limit);
+                let mut create_input = CreateInputs {
+                    caller,
+                    scheme,
+                    value,
+                    init_code: data,
+                    gas_limit,
+                };
+                let (exit, address, ret_gas, bytes) = self.create_inner::<GSPEC>(&mut create_input);
                 (exit, ret_gas, TransactOut::Create(bytes, address))
             }
         };
@@ -141,10 +150,9 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> Transact
         if crate::USE_GAS {
             gas.reimburse_unspend(&exit_reason, ret_gas);
         }
-        match self.finalize::<GSPEC>(caller, &gas) {
-            Err(e) => (e, out, gas.spend(), Map::new(), Vec::new()),
-            Ok((state, logs)) => (exit_reason, out, gas.spend(), state, logs),
-        }
+
+        let (state, logs, gas_used) = self.finalize::<GSPEC>(caller, &gas);
+        (exit_reason, out, gas_used, state, logs)
     }
 }
 
@@ -186,9 +194,9 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
         &mut self,
         caller: H160,
         gas: &Gas,
-    ) -> Result<(Map<H160, Account>, Vec<Log>), Return> {
+    ) -> (Map<H160, Account>, Vec<Log>, u64) {
         let coinbase = self.data.env.block.coinbase;
-        if crate::USE_GAS {
+        let gas_used = if crate::USE_GAS {
             let effective_gas_price = self.data.env.effective_gas_price();
             let basefee = self.data.env.block.basefee;
             let max_refund_quotient = if SPEC::enabled(LONDON) { 5 } else { 2 }; // EIP-3529: Reduction in refunds
@@ -207,11 +215,13 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
             self.data
                 .subroutine
                 .balance_add(coinbase, coinbase_gas_price * (gas.spend() - gas_refunded));
+            gas.spend() - gas_refunded
         } else {
             // touch coinbase
             self.data.subroutine.load_account(coinbase, self.data.db);
             self.data.subroutine.balance_add(coinbase, U256::zero());
-        }
+            0
+        };
         let (mut new_state, logs) = self.data.subroutine.finalize();
         // precompiles are special case. If there is precompiles in finalized Map that means some balance is
         // added to it, we need now to load precompile address from db and add this amount to it so that we
@@ -225,7 +235,7 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
             }
         }
 
-        Ok((new_state, logs))
+        (new_state, logs, gas_used)
     }
 
     fn inner_load_account(&mut self, caller: H160) -> bool {
@@ -285,41 +295,48 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
 
     fn create_inner<SPEC: Spec>(
         &mut self,
-        caller: H160,
-        scheme: CreateScheme,
-        value: U256,
-        init_code: Bytes,
-        gas_limit: u64,
+        inputs: &mut CreateInputs,
     ) -> (Return, Option<H160>, Gas, Bytes) {
-        let gas = Gas::new(gas_limit);
-        self.load_account(caller);
+        // Call inspector
+        if INSPECT {
+            let (ret, address, gas, out) = self.inspector.create(&mut self.data, inputs);
+            if ret != Return::Continue {
+                return self
+                    .inspector
+                    .create_end(&mut self.data, inputs, ret, address, gas, out);
+            }
+        }
 
-        // check depth of calls
-        if self.data.subroutine.depth() > machine::CALL_STACK_LIMIT {
+        let gas = Gas::new(inputs.gas_limit);
+        self.load_account(inputs.caller);
+
+        // Check depth of calls
+        if self.data.subroutine.depth() > interpreter::CALL_STACK_LIMIT {
             return (Return::CallTooDeep, None, gas, Bytes::new());
         }
-        // check balance of caller and value. Do this before increasing nonce
-        if self.balance(caller).0 < value {
+        // Check balance of caller and value. Do this before increasing nonce
+        if self.balance(inputs.caller).0 < inputs.value {
             return (Return::OutOfFund, None, gas, Bytes::new());
         }
 
-        // inc nonce of caller
-        let old_nonce = self.data.subroutine.inc_nonce(caller);
-        // create address
-        let code_hash = H256::from_slice(Keccak256::digest(&init_code).as_slice());
-        let created_address = match scheme {
-            CreateScheme::Create => util::create_address(caller, old_nonce),
-            CreateScheme::Create2 { salt } => util::create2_address(caller, code_hash, salt),
+        // Increase nonce of caller
+        let old_nonce = self.data.subroutine.inc_nonce(inputs.caller);
+
+        // Create address
+        let code_hash = H256::from_slice(Keccak256::digest(&inputs.init_code).as_slice());
+        let created_address = match inputs.scheme {
+            CreateScheme::Create => create_address(inputs.caller, old_nonce),
+            CreateScheme::Create2 { salt } => create2_address(inputs.caller, code_hash, salt),
         };
         let ret = Some(created_address);
 
-        // load account so that it will be hot
+        // Load account so that it will be hot
         self.load_account(created_address);
 
-        // enter into subroutine
+        // Enter subroutine
         let checkpoint = self.data.subroutine.create_checkpoint();
 
-        // create contract account and check for collision
+        // Create contract account and check for collision
         if !self.data.subroutine.new_contract_acc(
             created_address,
             self.precompiles.contains(&created_address),
@@ -329,39 +346,58 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
             return (Return::CreateCollision, ret, gas, Bytes::new());
         }
 
-        // transfer value to contract address
-        if let Err(e) = self
-            .data
-            .subroutine
-            .transfer(caller, created_address, value, self.data.db)
-        {
+        // Transfer value to contract address
+        if let Err(e) = self.data.subroutine.transfer(
+            inputs.caller,
+            created_address,
+            inputs.value,
+            self.data.db,
+        ) {
             self.data.subroutine.checkpoint_revert(checkpoint);
             return (e, ret, gas, Bytes::new());
         }
-        // inc nonce of contract
+
+        // Increase nonce of the contract
         if SPEC::enabled(ISTANBUL) {
             self.data.subroutine.inc_nonce(created_address);
         }
-        // create new machine and execute init function
-        let contract =
-            Contract::new::<SPEC>(Bytes::new(), init_code, created_address, caller, value);
-        let mut machine = Machine::new::<SPEC>(contract, gas.limit(), self.data.subroutine.depth());
+
+        // Create new interpreter and execute initcode
+        let contract = Contract::new::<SPEC>(
+            Bytes::new(),
+            inputs.init_code.clone(),
+            created_address,
+            inputs.caller,
+            inputs.value,
+        );
+
+        #[cfg(feature = "memory_limit")]
+        let mut interp = Interpreter::new_with_memory_limit::<SPEC>(
+            contract,
+            gas.limit(),
+            self.data.env.cfg.memory_limit,
+        );
+
+        #[cfg(not(feature = "memory_limit"))]
+        let mut interp = Interpreter::new::<SPEC>(contract, gas.limit());
+
         if Self::INSPECT {
             self.inspector
-                .initialize_machine(&mut machine, &mut self.data, false); // TODO fix is_static
+                .initialize_interp(&mut interp, &mut self.data, false); // TODO fix is_static
         }
-        let exit_reason = machine.run::<Self, SPEC>(self);
+        let exit_reason = interp.run::<Self, SPEC>(self);
+
         // Host error if present on execution\
-        let ret = match exit_reason {
+        let (ret, address, gas, out) = match exit_reason {
             return_ok!() => {
                 let b = Bytes::new();
                 // if ok, check contract creation limit and calculate gas deduction on output len.
-                let code = machine.return_value();
+                let code = interp.return_value();
 
                 // EIP-3541: Reject new contract code starting with the 0xEF byte
                 if SPEC::enabled(LONDON) && !code.is_empty() && code.get(0) == Some(&0xEF) {
                     self.data.subroutine.checkpoint_revert(checkpoint);
-                    return (Return::CreateContractWithEF, ret, machine.gas, b);
+                    return (Return::CreateContractWithEF, ret, interp.gas, b);
                 }
 
                 // TODO maybe create some macro to hide this `if`
@@ -375,14 +411,14 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
                 // EIP-170: Contract code size limit
                 if SPEC::enabled(SPURIOUS_DRAGON) && code.len() > contract_code_size_limit {
                     self.data.subroutine.checkpoint_revert(checkpoint);
-                    return (Return::CreateContractLimit, ret, machine.gas, b);
+                    return (Return::CreateContractLimit, ret, interp.gas, b);
                 }
                 if crate::USE_GAS {
-                    let gas_for_code = code.len() as u64 * crate::instructions::gas::CODEDEPOSIT;
+                    let gas_for_code = code.len() as u64 * crate::gas::CODEDEPOSIT;
                     // record code deposit gas cost and check if we are out of gas.
-                    if !machine.gas.record_cost(gas_for_code) {
+                    if !interp.gas.record_cost(gas_for_code) {
                         self.data.subroutine.checkpoint_revert(checkpoint);
-                        return (Return::OutOfGas, ret, machine.gas, b);
+                        return (Return::OutOfGas, ret, interp.gas, b);
                     }
                 }
                 // if we have enought gas
@@ -391,52 +427,65 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
                 self.data
                     .subroutine
                     .set_code(created_address, code, code_hash);
-                (Return::Continue, ret, machine.gas, b)
+                (Return::Continue, ret, interp.gas, b)
             }
             _ => {
                 self.data.subroutine.checkpoint_revert(checkpoint);
-                (exit_reason, ret, machine.gas, machine.return_value())
+                (exit_reason, ret, interp.gas, interp.return_value())
             }
         };
+
         if INSPECT {
-            self.inspector.call_end();
+            self.inspector
+                .create_end(&mut self.data, inputs, ret, address, gas, out)
+        } else {
+            (ret, address, gas, out)
         }
-        ret
     }
 
-    #[allow(clippy::too_many_arguments)]
-    fn call_inner<SPEC: Spec>(
-        &mut self,
-        code_address: H160,
-        transfer: Transfer,
-        input: Bytes,
-        gas_limit: u64,
-        context: CallContext,
-    ) -> (Return, Gas, Bytes) {
-        let mut gas = Gas::new(gas_limit);
-        // Load account and get code. Account is now hot.
-        let (code, _) = self.code(code_address);
+    fn call_inner<SPEC: Spec>(&mut self, inputs: &mut CallInputs) -> (Return, Gas, Bytes) {
+        // Call the inspector
+        if INSPECT {
+            let (ret, gas, out) = self
+                .inspector
+                .call(&mut self.data, inputs, SPEC::IS_STATIC_CALL);
+            if ret != Return::Continue {
+                return self.inspector.call_end(
+                    &mut self.data,
+                    inputs,
+                    gas,
+                    ret,
+                    out,
+                    SPEC::IS_STATIC_CALL,
+                );
+            }
+        }
 
-        // check depth
-        if self.data.subroutine.depth() > machine::CALL_STACK_LIMIT {
+        let mut gas = Gas::new(inputs.gas_limit);
+        // Load account and get code. Account is now hot.
+        let (code, _) = self.code(inputs.contract);
+
+        // Check depth
+        if self.data.subroutine.depth() > interpreter::CALL_STACK_LIMIT {
             return (Return::CallTooDeep, gas, Bytes::new());
         }
 
         // Create subroutine checkpoint
         let checkpoint = self.data.subroutine.create_checkpoint();
-        // touch address. For "EIP-158 State Clear" this will erase empty accounts.
-        if transfer.value.is_zero() {
-            self.load_account(context.address);
+
+        // Touch address. For "EIP-158 State Clear", this will erase empty accounts.
+        if inputs.transfer.value.is_zero() {
+            self.load_account(inputs.context.address);
             self.data
                 .subroutine
-                .balance_add(context.address, U256::zero()); // touch the acc
+                .balance_add(inputs.context.address, U256::zero());
         }
 
-        // transfer value from caller to called account;
+        // Transfer value from caller to called account
         match self.data.subroutine.transfer(
-            transfer.source,
-            transfer.target,
-            transfer.value,
+            inputs.transfer.source,
+            inputs.transfer.target,
+            inputs.transfer.value,
             self.data.db,
         ) {
             Err(e) => {
@@ -446,11 +495,11 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
             Ok((_source_is_cold, _target_is_cold)) => {}
         }
 
-        // call precompiles
-        if let Some(precompile) = self.precompiles.get(&code_address) {
+        // Call precompiles
+        let (ret, gas, out) = if let Some(precompile) = self.precompiles.get(&inputs.contract) {
             let out = match precompile {
-                Precompile::Standard(fun) => fun(input.as_ref(), gas_limit),
-                Precompile::Custom(fun) => fun(input.as_ref(), gas_limit),
+                Precompile::Standard(fun) => fun(inputs.input.as_ref(), inputs.gas_limit),
+                Precompile::Custom(fun) => fun(inputs.input.as_ref(), inputs.gas_limit),
             };
             match out {
                 Ok(PrecompileOutput { output, cost, logs }) => {
@@ -475,22 +524,39 @@ impl<'a, GSPEC: Spec, DB: Database, const INSPECT: bool> EVMImpl<'a, GSPEC, DB,
                 }
             }
         } else {
-            // create machine and execute subcall
-            let contract = Contract::new_with_context::<SPEC>(input, code, &context);
-            let mut machine =
-                Machine::new::<SPEC>(contract, gas_limit, self.data.subroutine.depth());
+            // Create interpreter and execute subcall
+            let contract =
+                Contract::new_with_context::<SPEC>(inputs.input.clone(), code, &inputs.context);
+
+            #[cfg(feature = "memory_limit")]
+            let mut interp = Interpreter::new_with_memory_limit::<SPEC>(
+                contract,
+                gas.limit(),
+                self.data.env.cfg.memory_limit,
+            );
+
+            #[cfg(not(feature = "memory_limit"))]
+            let mut interp = Interpreter::new::<SPEC>(contract, gas.limit());
+
             if Self::INSPECT {
                 self.inspector
-                    .initialize_machine(&mut machine, &mut self.data, false); // TODO fix is_static
+                    .initialize_interp(&mut interp, &mut self.data, false); // TODO fix is_static
             }
-            let exit_reason = machine.run::<Self, SPEC>(self);
+            let exit_reason = interp.run::<Self, SPEC>(self);
             if matches!(exit_reason, return_ok!()) {
                 self.data.subroutine.checkpoint_commit();
             } else {
                 self.data.subroutine.checkpoint_revert(checkpoint);
             }
 
-            (exit_reason, machine.gas, machine.return_value())
+            (exit_reason, interp.gas, interp.return_value())
+        };
+
+        if INSPECT {
+            self.inspector
+                .call_end(&mut self.data, inputs, gas, ret, out, SPEC::IS_STATIC_CALL)
+        } else {
+            (ret, gas, out)
         }
     }
 }
@@ -501,13 +567,13 @@ impl<'a, GSPEC: Spec, DB: Database + 'a, const INSPECT: bool> Host
     const INSPECT: bool = INSPECT;
     type DB = DB;
 
-    fn step(&mut self, machine: &mut Machine, is_static: bool) -> Return {
-        self.inspector.step(machine, &mut self.data, is_static);
-        Return::Continue
+    fn step(&mut self, interp: &mut Interpreter, is_static: bool) -> Return {
+        self.inspector.step(interp, &mut self.data, is_static)
     }
 
-    fn step_end(&mut self, _ret: Return, _machine: &mut Machine) -> Return {
-        Return::Continue
+    fn step_end(&mut self, interp: &mut Interpreter, is_static: bool, ret: Return) -> Return {
+        self.inspector
+            .step_end(interp, &mut self.data, is_static, ret)
     }
 
     fn env(&mut self) -> &mut Env {
@@ -563,6 +629,9 @@ impl<'a, GSPEC: Spec, DB: Database + 'a, const INSPECT: bool> Host
     }
 
     fn log(&mut self, address: H160, topics: Vec<H256>, data: Bytes) {
+        if INSPECT {
+            self.inspector.log(&mut self.data, &address, &topics, &data);
+        }
         let log = Log {
             address,
             topics,
@@ -582,57 +651,46 @@ impl<'a, GSPEC: Spec, DB: Database + 'a, const INSPECT: bool> Host
 
     fn create<SPEC: Spec>(
         &mut self,
-        caller: H160,
-        scheme: CreateScheme,
-        value: U256,
-        init_code: Bytes,
-        gas: u64,
+        inputs: &mut CreateInputs,
     ) -> (Return, Option<H160>, Gas, Bytes) {
-        if INSPECT {
-            let (ret, gas, bytes, out) =
-                self.inspector
-                    .create(&mut self.data, caller, &scheme, value, &init_code, gas);
-            if ret != Return::Continue {
-                return (ret, gas, bytes, out);
-            }
-        }
-        self.create_inner::<SPEC>(caller, scheme, value, init_code, gas)
+        self.create_inner::<SPEC>(inputs)
     }
 
-    fn call<SPEC: Spec>(
-        &mut self,
-        code_address: H160,
-        transfer: Transfer,
-        input: Bytes,
-        gas: u64,
-        context: CallContext,
-    ) -> (Return, Gas, Bytes) {
-        if INSPECT {
-            let (ret, gas, bytes) = self.inspector.call(
-                &mut self.data,
-                code_address,
-                &context,
-                &transfer,
-                &input,
-                gas,
-                SPEC::IS_STATIC_CALL,
-            );
-            if ret != Return::Continue {
-                return (ret, gas, bytes);
-            }
-        }
-        self.call_inner::<SPEC>(code_address, transfer, input, gas, context)
+    fn call<SPEC: Spec>(&mut self, inputs: &mut CallInputs) -> (Return, Gas, Bytes) {
+        self.call_inner::<SPEC>(inputs)
     }
 }
 
+pub fn create_address(caller: H160, nonce: u64) -> H160 {
+    let mut stream = rlp::RlpStream::new_list(2);
+    stream.append(&caller);
+    stream.append(&nonce);
+    let out = H256::from_slice(Keccak256::digest(&stream.out()).as_slice());
+    let out = H160::from_slice(&out.as_bytes()[12..]);
+    out
+}
+
+/// Get the create address from given scheme.
+pub fn create2_address(caller: H160, code_hash: H256, salt: U256) -> H160 {
+    let mut temp: [u8; 32] = [0; 32];
+    salt.to_big_endian(&mut temp);
+
+    let mut hasher = Keccak256::new();
+    hasher.update(&[0xff]);
+    hasher.update(&caller[..]);
+    hasher.update(&temp);
+    hasher.update(&code_hash[..]);
+    H160::from_slice(&hasher.finalize().as_slice()[12..])
+}
+
 /// EVM context host.
 pub trait Host {
     const INSPECT: bool;
 
     type DB: Database;
 
-    fn step(&mut self, machine: &mut Machine, is_static: bool) -> Return;
-    fn step_end(&mut self, ret: Return, machine: &mut Machine) -> Return;
+    fn step(&mut self, interp: &mut Interpreter, is_static: bool) -> Return;
+    fn step_end(&mut self, interp: &mut Interpreter, is_static: bool, ret: Return) -> Return;
 
     fn env(&mut self) -> &mut Env;
 
@@ -657,20 +715,8 @@ pub trait Host {
     /// Invoke a create operation.
     fn create<SPEC: Spec>(
         &mut self,
-        caller: H160,
-        scheme: CreateScheme,
-        value: U256,
-        init_code: Bytes,
-        gas: u64,
+        inputs: &mut CreateInputs,
     ) -> (Return, Option<H160>, Gas, Bytes);
-
     /// Invoke a call operation.
-    fn call<SPEC: Spec>(
-        &mut self,
-        code_address: H160,
-        transfer: Transfer,
-        input: Bytes,
-        gas: u64,
-        context: CallContext,
-    ) -> (Return, Gas, Bytes);
+    fn call<SPEC: Spec>(&mut self, input: &mut CallInputs) -> (Return, Gas, Bytes);
 }
