@@ -1,17 +1,16 @@
 #![allow(dead_code)]
-//#![forbid(unsafe_code, unused_variables, unused_imports)]
-//#![no_std] only blocker in auto_impl check: https://github.com/bluealloy/revm/issues/4
+//#![no_std]
 
 pub mod db;
 mod evm;
 mod evm_impl;
+pub(crate) mod gas;
 mod inspector;
 mod instructions;
-mod machine;
+mod interpreter;
 mod models;
-mod spec;
+mod specification;
 mod subroutine;
-mod util;
 
 pub use evm_impl::{EVMData, Host};
 
@@ -19,15 +18,16 @@ pub type DummyStateDB = InMemoryDB;
 
 pub use db::{Database, DatabaseCommit, InMemoryDB};
 pub use evm::{new, EVM};
+pub use gas::Gas;
 pub use inspector::{Inspector, NoOpInspector, OverrideSpec};
 pub use instructions::{
     opcode::{self, spec_opcode_gas, OpCode, OPCODE_JUMPMAP},
     Return,
 };
-pub use machine::{Gas, Machine};
+pub use interpreter::{Contract, Interpreter, Memory, Stack};
 pub use models::*;
-pub use spec::*;
-pub use subroutine::{Account, SubRoutine};
+pub use specification::*;
+pub use subroutine::{Account, Filth, SubRoutine};
 
 extern crate alloc;
 
